"""Mixin classes used to inherit parameter functionality."""

import typing

from . import engines
from . import formats
from . import renderers
from . import formatters

__all__ = ['Parameters']


class Parameters(engines.Engine, formats.Format,
                 renderers.Renderer, formatters.Formatter):
    """Parameters for calling ``graphviz.render()`` and ``graphviz.pipe()``."""

    def _get_parameters(self, *,
                        engine: typing.Optional[str] = None,
                        format: typing.Optional[str] = None,
                        renderer: typing.Optional[str] = None,
                        formatter: typing.Optional[str] = None,
                        verify: bool = False,
                        **kwargs):
        if engine is None:
            engine = self.engine
        elif verify:
            self._verify_engine(engine)

        if format is None:
            format = self.format
        elif verify:
            self._verify_format(format)

        if renderer is None:
            renderer = self.renderer
        elif verify:
            self._verify_renderer(renderer)

        if formatter is None:
            formatter = self.formatter
        elif verify:
            self._verify_formatter(formatter)

        kwargs.update(engine=engine, format=format,
                      renderer=renderer, formatter=formatter)
        return kwargs



"""Rendering parameter handling."""

from .. import copying

__all__ = ['ParameterBase']


class ParameterBase(copying.CopyBase):
    """Rendering parameter."""

    def _getattr_from_dict(self, attrname: str, *, default=None):
        """Return self.attrname if attrname is in the instance dictionary
            (as oposed to on the type)."""
        if attrname in self.__dict__:
            return getattr(self, attrname)
        return default



"""Open files in platform-specific default viewing application."""

import logging
import os
import platform
import subprocess
import typing

from .. import _tools

__all__ = ['view']

PLATFORM = platform.system().lower()


log = logging.getLogger(__name__)


@_tools.deprecate_positional_args(supported_number=1)
def view(filepath: typing.Union[os.PathLike, str],
         quiet: bool = False) -> None:
    """Open filepath with its default viewing application (platform-specific).

    Args:
        filepath: Path to the file to open in viewer.
        quiet: Suppress ``stderr`` output
            from the viewer process (ineffective on Windows).

    Raises:
        RuntimeError: If the current platform is not supported.

    Note:
        There is no option to wait for the application to close,
        and no way to retrieve the application's exit status.
    """
    try:
        view_func = getattr(view, PLATFORM)
    except AttributeError:
        raise RuntimeError(f'platform {PLATFORM!r} not supported')
    view_func(filepath, quiet=quiet)


@_tools.attach(view, 'darwin')
def view_darwin(filepath: typing.Union[os.PathLike, str], *,
                quiet: bool) -> None:
    """Open filepath with its default application (mac)."""
    cmd = ['open', filepath]
    log.debug('view: %r', cmd)
    kwargs = {'stderr': subprocess.DEVNULL} if quiet else {}
    subprocess.Popen(cmd, **kwargs)


@_tools.attach(view, 'linux')
@_tools.attach(view, 'freebsd')
def view_unixoid(filepath: typing.Union[os.PathLike, str], *,
                 quiet: bool) -> None:
    """Open filepath in the user's preferred application (linux, freebsd)."""
    cmd = ['xdg-open', filepath]
    log.debug('view: %r', cmd)
    kwargs = {'stderr': subprocess.DEVNULL} if quiet else {}
    subprocess.Popen(cmd, **kwargs)


@_tools.attach(view, 'windows')
def view_windows(filepath: typing.Union[os.PathLike, str], *,
                 quiet: bool) -> None:
    """Start filepath with its associated application (windows)."""
    # TODO: implement quiet=True
    filepath = os.path.normpath(filepath)
    log.debug('view: %r', filepath)
    os.startfile(filepath)  # pytype: disable=module-attr



"""Python 3.8 compatibility and platform compatibility."""

import platform
import sys

if sys.version_info < (3, 9):  # pragma: no cover
    # pytype not supported
    import unittest.mock

    Literal = unittest.mock.MagicMock(name='Literal')
else:  # pragma: no cover
    from typing import Literal

    Literal = Literal  # CAVEAT: use None instead of Literal[None]


def get_startupinfo() -> None:
    """Return None for startupinfo argument of ``subprocess.Popen``."""
    return None


assert get_startupinfo() is None, 'get_startupinfo() defaults to a no-op'


if platform.system() == 'Windows':  # pragma: no cover
    import subprocess

    def get_startupinfo() -> subprocess.STARTUPINFO:  # pytype: disable=module-attr
        """Return subprocess.STARTUPINFO instance hiding the console window."""
        startupinfo = subprocess.STARTUPINFO()  # pytype: disable=module-attr
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # pytype: disable=module-attr
        startupinfo.wShowWindow = subprocess.SW_HIDE  # pytype: disable=module-attr
        return startupinfo



"""Display rendered graph as SVG in Jupyter Notebooks and QtConsole."""

import typing

from . import piping

__all__ = ['JUPYTER_FORMATS',
           'SUPPORTED_JUPYTER_FORMATS', 'DEFAULT_JUPYTER_FORMAT',
           'get_jupyter_format_mimetype',
           'JupyterIntegration']

_IMAGE_JPEG = 'image/jpeg'

JUPYTER_FORMATS = {'jpeg': _IMAGE_JPEG,
                   'jpg': _IMAGE_JPEG,
                   'png': 'image/png',
                   'svg': 'image/svg+xml'}

SUPPORTED_JUPYTER_FORMATS = set(JUPYTER_FORMATS)

DEFAULT_JUPYTER_FORMAT = next(_ for _ in SUPPORTED_JUPYTER_FORMATS if _ == 'svg')

MIME_TYPES = {'image/jpeg': '_repr_image_jpeg',
              'image/png': '_repr_image_png',
              'image/svg+xml': '_repr_image_svg_xml'}

assert MIME_TYPES.keys() == set(JUPYTER_FORMATS.values())

SVG_ENCODING = 'utf-8'


def get_jupyter_format_mimetype(jupyter_format: str) -> str:
    try:
        return JUPYTER_FORMATS[jupyter_format]
    except KeyError:
        raise ValueError(f'unknown jupyter_format: {jupyter_format!r}'
                         f' (must be one of {sorted(JUPYTER_FORMATS)})')


def get_jupyter_mimetype_format(mimetype: str) -> str:
    if mimetype not in MIME_TYPES:
        raise ValueError(f'unsupported mimetype: {mimetype!r}'
                         f' (must be one of {sorted(MIME_TYPES)})')

    assert mimetype in JUPYTER_FORMATS.values()

    for format, jupyter_mimetype in JUPYTER_FORMATS.items():
        if jupyter_mimetype == mimetype:
            return format

    raise RuntimeError  # pragma: no cover


class JupyterIntegration(piping.Pipe):
    """Display rendered graph as SVG in Jupyter Notebooks and QtConsole."""

    _jupyter_mimetype = get_jupyter_format_mimetype(DEFAULT_JUPYTER_FORMAT)

    def _repr_mimebundle_(self,
                          include: typing.Optional[typing.Iterable[str]] = None,
                          exclude: typing.Optional[typing.Iterable[str]] = None,
                          **_) -> typing.Dict[str, typing.Union[bytes, str]]:
        r"""Return the rendered graph as IPython mimebundle.

        Args:
            include: Iterable of mimetypes to include in the result.
                If not given or ``None``: ``['image/sxg+xml']``.
            exclude: Iterable of minetypes to exclude from the result.
                Overrides ``include``.

        Returns:
            Mapping from mimetypes to data.

        Example:
            >>> doctest_mark_exe()
            >>> import graphviz
            >>> dot = graphviz.Graph()
            >>> dot._repr_mimebundle_()  # doctest: +ELLIPSIS
            {'image/svg+xml': '<?xml version=...
            >>> dot._repr_mimebundle_(include=['image/png'])  # doctest: +ELLIPSIS
            {'image/png': b'\x89PNG...
            >>> dot._repr_mimebundle_(include=[])
            {}
            >>> dot._repr_mimebundle_(include=['image/svg+xml', 'image/jpeg'],
            ...                       exclude=['image/svg+xml'])  # doctest: +ELLIPSIS
            {'image/jpeg': b'\xff...
            >>> list(dot._repr_mimebundle_(include=['image/png', 'image/jpeg']))
            ['image/jpeg', 'image/png']

        See also:
            IPython documentation:
            - https://ipython.readthedocs.io/en/stable/api/generated/IPython.display.html#functions
            - https://ipython.readthedocs.io/en/stable/config/integrating.html#MyObject._repr_mimebundle_  # noqa: E501
            - https://nbviewer.org/github/ipython/ipython/blob/master/examples/IPython%20Kernel/Custom%20Display%20Logic.ipynb#Custom-Mimetypes-with-_repr_mimebundle_  # noqa: E501
        """
        include = set(include) if include is not None else {self._jupyter_mimetype}
        include -= set(exclude or [])
        return {mimetype: getattr(self, method_name)()
                for mimetype, method_name in MIME_TYPES.items()
                if mimetype in include}

    def _repr_image_jpeg(self) -> bytes:
        """Return the rendered graph as JPEG bytes."""
        return self.pipe(format='jpeg')

    def _repr_image_png(self) -> bytes:
        """Return the rendered graph as PNG bytes."""
        return self.pipe(format='png')

    def _repr_image_svg_xml(self) -> str:
        """Return the rendered graph as SVG string."""
        return self.pipe(format='svg', encoding=SVG_ENCODING)



"""Rendering engine parameter handling."""

import typing

from . import base

__all__ = ['ENGINES', 'verify_engine', 'Engine']

ENGINES = {'dot',  # https://www.graphviz.org/pdf/dot.1.pdf
           'neato',
           'twopi',
           'circo',
           'fdp',
           'sfdp',
           'patchwork',
           'osage'}

DEFAULT_ENGINE = 'dot'

REQUIRED = True


def verify_engine(engine: str, *, required: bool = REQUIRED) -> None:
    if engine is None:
        if required:
            raise ValueError('missing engine')
    elif engine.lower() not in ENGINES:
        raise ValueError(f'unknown engine: {engine!r}'
                         f' (must be one of {sorted(ENGINES)})')


class Engine(base.ParameterBase):
    """Rendering engine parameter with ``'dot''`` default."""

    _engine = DEFAULT_ENGINE

    _verify_engine = staticmethod(verify_engine)

    def __init__(self, *, engine: typing.Optional[str] = None, **kwargs) -> None:
        super().__init__(**kwargs)

        if engine is not None:
            self.engine = engine

    def _copy_kwargs(self, **kwargs):
        """Return the kwargs to create a copy of the instance."""
        engine = self._getattr_from_dict('_engine')
        if engine is not None:
            kwargs['engine'] = engine
        return super()._copy_kwargs(**kwargs)

    @property
    def engine(self) -> str:
        """The layout engine used for rendering
            (``'dot'``, ``'neato'``, ...)."""
        return self._engine

    @engine.setter
    def engine(self, engine: str) -> None:
        engine = engine.lower()
        self._verify_engine(engine)
        self._engine = engine



"""Commonly used exception classes."""

from .backend.execute import ExecutableNotFound, CalledProcessError

__all__ = ['ExecutableNotFound', 'CalledProcessError',
           'RequiredArgumentError', 'FileExistsError',
           'UnknownSuffixWarning', 'FormatSuffixMismatchWarning',
           'DotSyntaxWarning']


class RequiredArgumentError(TypeError):
    """:exc:`TypeError` raised if a required argument is missing."""


class FileExistsError(FileExistsError):
    """:exc:`FileExistsError` raised with ``raise_if_exists=True``."""


class UnknownSuffixWarning(RuntimeWarning):
    """:exc:`RuntimeWarning` raised if the suffix of ``outfile`` is unknown
        and the given ``format`` is used instead."""


class FormatSuffixMismatchWarning(UserWarning):
    """:exc:`UserWarning` raised if the suffix ``outfile``
        does not match the given ``format``."""


class DotSyntaxWarning(RuntimeWarning):
    """:exc:`RuntimeWarning` raised if a quoted string
        is expected to cause a ``CalledProcessError`` from rendering."""
